import json
from pathlib import Path
from sc2.constants import *
import random
import sc2
import numpy as np
from sc2 import Race
from sc2.position import Point2, Point3


# Bots are created as classes and they need to have on_step method defined.
# Do not change the name of the class!
class MyBot(sc2.BotAI):
    with open(Path(__file__).parent / "botinfo.json") as f:
        NAME = json.load(f)["name"]

    retreat_pos = None
    units_to_ignore = [EGG, LARVA, OVERLORD, OVERSEER, OBSERVER, BROODLING, INTERCEPTOR, MEDIVAC, CREEPTUMOR, CREEPTUMORBURROWED, CREEPTUMORQUEEN, CREEPTUMORMISSILE]
    has_sent_workers = False
    under_construction = {}
    remembered_friendly_units_by_tag = {}
    remembered_enemy_units_by_tag = {}
    remembered_enemy_units = []
    ready_to_fight   = False
    random_exp_location = None
    visited_locations = []

    # On_step method is invoked each game-tick and should not take more than
    # 2 seconds to run, otherwise the bot will timeout and cannot receive new
    # orders.
    # It is important to note that on_step is asynchronous - meaning practices
    # for asynchronous programming should be followed.
    async def scout(self):
        scout = None

        # Check if we already have a scout (a worker with PATROL order)
        for worker in self.workers:
            if self.has_order([PATROL], worker):
                scout = worker

        # If we don't have a scout, select one, and order it to move to random exp
        if not scout:
            random_exp_location = random.choice(list(self.expansion_locations.keys()))
            scout = self.workers.closest_to(self.start_location)

            if not scout:
                return

            await self.order(scout, PATROL, random_exp_location)
            return

        # Basic avoidance: If enemy is too close, go to map center
        nearby_enemy_units = self.known_enemy_units.filter(lambda unit: unit.type_id not in self.units_to_ignore).closer_than(10, scout)
        if nearby_enemy_units.exists:
            await self.order(scout, PATROL, self.game_info.map_center)
            return

        # We're close enough, so change target
        target = sc2.position.Point2((scout.orders[0].target.x, scout.orders[0].target.y))
        if scout.distance_to(target) < 10:
            random_exp_location = random.choice(list(self.expansion_locations.keys()))
            await self.order(scout, PATROL, random_exp_location)
            return

    def remember_enemy_units(self):
        # Every 60 seconds, clear all remembered units (to clear out killed units)
        #if round(self.get_game_time() % 60) == 0:
        #    self.remembered_enemy_units_by_tag = {}

        # Look through all currently seen units and add them to list of remembered units (override existing)
        for unit in self.known_enemy_units:
            unit.is_known_this_step = True
            self.remembered_enemy_units_by_tag[unit.tag] = unit

        # Convert to an sc2 Units object and place it in self.remembered_enemy_units
        self.remembered_enemy_units = sc2.units.Units([], self._game_data)
        for tag, unit in list(self.remembered_enemy_units_by_tag.items()):
            # Make unit.is_seen = unit.is_visible 
            if unit.is_known_this_step:
                unit.is_seen = unit.is_visible # There are known structures that are not visible
                unit.is_known_this_step = False # Set to false for next step
            else:
                unit.is_seen = False

            # Units that are not visible while we have friendly units nearby likely don't exist anymore, so delete them
            if not unit.is_seen and self.units.closer_than(7, unit).exists:
                del self.remembered_enemy_units_by_tag[tag]
                continue

            self.remembered_enemy_units.append(unit)

    def enemy_army_value(self, position, distance=10):
        value = 0

        for unit in self.known_enemy_units.ready.not_structure.closer_than(distance, position):
            value += unit.health + unit.shield

            # Add extra army value for marine/marauder, to not under-estimate
            if unit.type_id in [MARINE, MARAUDER]:
                value += 20

        # Count nearby cannons
        for unit in self.known_enemy_units(PHOTONCANNON).ready.closer_than(10, position):
            value += unit.health # Skip shield, to not overestimate

        # Count nearby bunkers
        for unit in self.known_enemy_units(BUNKER).ready.closer_than(10, position):
            value += unit.health

        # Count nearby spine crawlers
        for unit in self.known_enemy_units(SPINECRAWLER).ready.closer_than(10, position):
            value += unit.health

        return value

    async def worker_split(self):
        for worker in self.workers:
            closest_mineral_patch = self.state.mineral_field.closest_to(worker)
            await self.do(worker.gather(closest_mineral_patch))
            #await self.order(worker, HARVEST_GATHER, closest_mineral_patch)    

    def get_game_time(self):
        return self.state.game_loop*0.725*(1/16)

    def remember_friendly_units(self):
        for unit in self.units:
            unit.is_taking_damage = False

            # If we already remember this friendly unit
            if unit.tag in self.remembered_friendly_units_by_tag:
                health_old = self.remembered_friendly_units_by_tag[unit.tag].health
                shield_old = self.remembered_friendly_units_by_tag[unit.tag].shield

                # Compare its health/shield since last step, to find out if it has taken any damage
                if unit.health < health_old or unit.shield < shield_old:
                    unit.is_taking_damage = True
                
            self.remembered_friendly_units_by_tag[unit.tag] = unit
    
    async def cancel_buildings(self):
        # Loop through all buildings that are under construction
        for building in self.units.structure.not_ready:
            # If we're not tracking this building, make sure to track it
            if building.tag not in self.under_construction:
                self.under_construction[building.tag] = {}
                self.under_construction[building.tag]["last_health"] = building.health

            # If health is low, and has dropped since last frame, cancel it!
            if building.health < 100 and building.health < self.under_construction[building.tag]["last_health"]:
                await self.do(building(CANCEL))
            else:
                self.under_construction[building.tag]["last_health"] = building.health    
    
    async def stalker_deffence(self, unit):		
        nearby_enemies = self.known_enemy_units.not_structure.filter(lambda unit: not unit.is_flying).closer_than(19, unit)		
        closestEnemies = nearby_enemies.closer_than(6, unit)
        if closestEnemies.amount > 0:
            if unit.weapon_cooldown == 0:
                await self.do(unit.attack(closestEnemies.random))
            elif unit.shield < 7:
                await self.do(unit.move(self.retreat_pos))    			                
        else:
            await self.do(unit.attack(self.main_base_ramp.depot_in_middle.towards(self.retreat_pos, 3)))

    async def stalker_attack(self, unit):	
        nearby_enemies = self.known_enemy_units.not_structure.filter(lambda unit: not unit.is_flying).closer_than(19, unit)
        closestEnemies = nearby_enemies.closer_than(6, unit)
        if closestEnemies.amount > 0:
            if unit.weapon_cooldown == 0:
                await self.do(unit.attack(closestEnemies.random))
            elif unit.shield < 7:
                await self.do(unit.move(self.retreat_pos))    			                
        else:
            await self.do(unit.attack(self.select_target(self.state)))          
                  
    async def select_worker(self, pos, force=False):
        if not self.workers.exists:
            return None
        #print("selected")
        # Find worker closest to pos, but make sure to only choose a worker that is gathering (to not cancel orders)
        worker = None
        for unit in self.workers.sorted_by_distance_to(pos):
            #if unit.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN, MOVE], unit):
            worker = unit
            break

        if worker is None:
            return None

        # Check if path is blocked
 
        return worker
       
        #distance = await self._client.query_pathing(worker.position, pos)
        #if distance is None:
        #    print("DISTANCE IS NONE!")
        #    # Path is blocked, so return random worker
        #    return self.workers.random
        #else:
        #    # Path not blocked
        #    return worker

    async def build(self, building, near, max_distance=20, unit=None, random_alternative=False, placement_step=2):
        """Build a building."""

        if isinstance(near, sc2.unit.Unit):
            near = near.position.to2
        elif near is not None:
            near = near.to2

        is_valid_location = False
        p = None

        p = await self.find_placement(building, near.rounded, max_distance, random_alternative, placement_step)
        if p is None:
            p = await self.find_placement(building, near.rounded, max_distance, True, placement_step)
            if p is None:
                return sc2.data.ActionResult.CantFindPlacementLocation

        unit = unit or await self.select_worker(p)

        if unit is None:
            return sc2.data.ActionResult.Error

        return await self.do(unit.build(building, p))


    async def order(self, units, order, target=None, silent=True):
        if type(units) != list:
            unit = units
            await self.do(unit(order, target=target))
        else:
            for unit in units:
                await self.do(unit(order, target=target))

    def has_order(self, orders, units):
        if type(orders) != list:
            orders = [orders]

        count = 0

        if type(units) == sc2.unit.Unit:
            unit = units
            if len(unit.orders) >= 1 and unit.orders[0].ability.id in orders:
                count += 1
        else:
            for unit in units:
                if len(unit.orders) >= 1 and unit.orders[0].ability.id in orders:
                  count += 1

        return count
    
    async def has_ability(self, ability, unit):
        abilities = await self.get_available_abilities(unit)
        if ability in abilities:
            return True
        else:
            return False
        
    def select_target(self, state):
        if self.known_enemy_units.structure.exists:
            return random.choice(self.known_enemy_units.structure).position
        
        elif self.time>450 :
            return self.random_exp_location
            
        return self.enemy_start_locations[0]
        
            
    async def on_step(self, iteration):
        
        actions = []
        if iteration == 0:
            #self._client.game_step = 1
            await self.chat_send(f"(cheese) (cheese) (pheart)")
            await self.worker_split()

        #for unit in self.units.not_structure:
        #    if not (unit.is_idle or self.has_order([MOVE], unit)):
        #        print(unit.orders[0].target)
        #for worker in self.workers:
        #    if not (worker.is_idle or self.has_order([MOVE], worker)):
        #        print(worker.orders)
        await self.distribute_workers()       
        scout = None
        await self.cancel_buildings()
        self.remember_friendly_units()

        nexus = None
    
        if not self.units(NEXUS).ready.exists:
            for unit in self.units.not_structure:
                await self.do(unit.attack(self.select_target(self.state)))
            return
        else:
            nexus = self.units(NEXUS).ready.random
        self.retreat_pos = nexus.position
        if int(self.time)%25==1:
            self.random_exp_location = random.choice(list(self.expansion_locations.keys())).position

        nearby_enemies = self.known_enemy_units.not_structure.filter(lambda unit: not unit.is_flying).closer_than(19, nexus)
        if nearby_enemies.amount >= 1 and self.workers.exists:
            units = self.units.not_structure.sorted_by_distance_to(nearby_enemies.first).take(nearby_enemies.amount*2, False)
            for unit in units:
                await self.do(unit.attack(nearby_enemies.closer_than(19, nexus).closest_to(unit)))
        else:
            for worker in self.workers.closer_than(50, nexus):
                if len(worker.orders) == 1 and worker.orders[0].ability.id in [ATTACK]:
                    await self.do(worker.gather(self.state.mineral_field.closest_to(nexus)))


            if self.workers.idle.closer_than(70, nexus).exists:
                for idle_worker in self.workers.idle.closer_than(70, nexus):
                    await self.do(idle_worker.gather(self.state.mineral_field.closest_to(nexus)))

        if self.workers.exists and self.units(VOIDRAY).amount >= 3:
            await self.scout()

        #if self.time>300 and not self.known_enemy_units.structure.exists and len(self.expansion_locations)>len(self.visited_locations):
        #    for location in self.expansion_locations:
        #        if location not in self.visited_locations:
        #            builder = await self.select_worker(location)
        #            if builder.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN, MOVE], builder):
        #                self.visited_locations.append(location)
        #                await self.do(builder.move(location))
           
        if (not self.has_sent_workers) and (self.already_pending(PYLON)) and (self.enemy_race==Race.Protoss or self.enemy_race==Race.Terran):
            self.has_sent_workers = True
            scout = await self.select_worker(self.enemy_start_locations[0])
            await self.do(scout.move(self.enemy_start_locations[0]))

        for worker in self.workers.closer_than(40, self.enemy_start_locations[0]):
            if (worker.is_idle or self.has_order([MOVE], worker) or self.units(ASSIMILATOR).closer_than(1.0, worker.position).exists) and self.known_enemy_units.closer_than(4, worker).not_structure.exists:
                x = 0
                y = 0
                k = 0
                for unit in self.known_enemy_units.closer_than(4, worker).not_structure:
                    x+=unit.position[0]
                    y+=unit.position[1]
                    k+=1
                x = x/k
                y = y/k
                #print(x, y, worker.position)
                #target = worker.position
                #target[0]+=target[0]-x
                #target[1]+=target[1]-y
                await self.do(worker.move(worker.position.towards(2*worker.position-Point2([x,y]), 4)))
        
            if self.has_sent_workers and self.already_pending(GATEWAY) and not self.units(GATEWAY).ready:
                vgs = self.state.vespene_geyser.closer_than(20.0, self.enemy_start_locations[0])
                for vg in vgs:
                    if not self.can_afford(ASSIMILATOR):
                        break

                #worker = self.select_build_worker(vg.position)
                    if worker is None:
                        break

                    if not self.units(ASSIMILATOR).closer_than(1.0, vg).exists and not self.known_enemy_units.structure.closer_than(1.0, vg).exists:
                        await self.do(worker.build(ASSIMILATOR, vg))

       # if self.enemy_race==Race.Zerg:
        if self.units(CYBERNETICSCORE).ready.exists:
            for gw in self.units(GATEWAY).ready:
                if gw.is_idle and self.can_afford(STALKER):
                    await self.do(gw.train(STALKER))                
            
        if self.already_pending(STARGATE) and not self.already_pending(SHIELDBATTERY) and self.can_afford(SHIELDBATTERY):
            builder = await self.select_worker(nexus.position)
            if builder is not None and (builder.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN, MOVE], builder)):
                gw = self.units(GATEWAY).closer_than(20.0, nexus)
                if gw.exists:
                    await self.build(SHIELDBATTERY, unit=builder, near=gw.random)

        if self.units(VOIDRAY).amount >= 1 and self.ready_to_fight is False:
            await self.chat_send("First VOIDRAY Ready"+str(self.time))
            print("First VOIDRAY Ready "+str(self.time))
            self.ready_to_fight = True            

        for sg in self.units(STARGATE).ready:
            if sg.is_idle and self.can_afford(VOIDRAY):
                await self.do(sg.train(VOIDRAY))
            if not sg.has_buff(BuffId.CHRONOBOOSTENERGYCOST) and not sg.is_idle:
                abilities = await self.get_available_abilities(nexus)
                if AbilityId.EFFECT_CHRONOBOOSTENERGYCOST in abilities:
                    await self.do(nexus(AbilityId.EFFECT_CHRONOBOOSTENERGYCOST, sg))
 
        if self.time>450:
            if self.can_afford(NEXUS) and self.units(NEXUS).amount < 6 :
                await self.expand_now()
            if self.can_afford(STARGATE) and not self.can_afford(NEXUS) and self.can_afford(VOIDRAY):
                builder = await self.select_worker(nexus.position)
                if builder is not None and (builder.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN, MOVE], builder)):
                    await self.build(STARGATE, unit=builder, near=nexus)
                   
        for sb in self.units(SHIELDBATTERY).ready:
             abilities = await self.get_available_abilities(sb)
             if AbilityId.EFFECT_RESTORE in abilities:
                 for unit in self.units.closer_than(6, sb.position):
                     if unit.shield < 5:
                         await self.do(sb(AbilityId.EFFECT_RESTORE, unit))
                     break
                                
        if self.units(STALKER).amount >= 1 and self.units(STALKER).amount <= int(self.time/100):
            for sk in self.units(STALKER):
                await self.stalker_deffence(sk)               

        if self.units(STALKER).amount > int(self.time/100):
            for sk in self.units(STALKER):
                await self.stalker_attack(sk)

        if self.units(VOIDRAY).amount >= 1 and self.ready_to_fight:
            for vr in self.units(VOIDRAY):
                if vr.shield > 40:#31:               
                    await self.do(vr.attack(self.select_target(self.state)))
                    nearby_enemy_units = self.known_enemy_units.not_structure.closer_than(15, vr)
                    if nearby_enemy_units.amount >= 1:
                        enemy_army_value = self.enemy_army_value(self.known_enemy_units.closest_to(vr), 10) #30
                        if self.has_ability(EFFECT_VOIDRAYPRISMATICALIGNMENT, vr) and enemy_army_value > 50 and not vr.has_buff(BuffId.VOIDRAYSWARMDAMAGEBOOST):
                            if await self.can_cast(vr, EFFECT_VOIDRAYPRISMATICALIGNMENT):
                                await self.order(vr, EFFECT_VOIDRAYPRISMATICALIGNMENT)                    
                elif self.units(SHIELDBATTERY).exists:
                    battery = None
                    for a in self.units(SHIELDBATTERY).sorted_by_distance_to(vr):
                        #if self.units(PYLON).closer_than(50, self.enemy_start_locations[0]).exists:
                        #    print("moving to objective", vr.position, a.position, self.units(PYLON).closer_than(50, self.enemy_start_locations[0]).random.position)
                        battery = a
                        break;
                    await self.do(vr.move(battery.position))                    
                elif self.units(PYLON).closer_than(50, self.enemy_start_locations[0]).exists:
                    await self.do(vr.move(self.units(PYLON).closer_than(50, self.enemy_start_locations[0]).random.position))

        for a in self.units(ASSIMILATOR).further_than(30, self.enemy_start_locations[0]):
            if a.assigned_harvesters < a.ideal_harvesters and len(self.workers)>10 and self.vespene<350:
                w = self.workers.closer_than(20, a)
                if w.exists:
                    await self.do(w.random.gather(a))
            if a.assigned_harvesters > a.ideal_harvesters or self.vespene>350:
                w = self.workers.closer_than(3, a)
                if w.exists:
                    await self.do(w.random.gather(self.state.mineral_field.closest_to(nexus)))

        if self.workers.amount < self.units(NEXUS).amount*20 and nexus.is_idle:
            if self.can_afford(PROBE):
                await self.do(nexus.train(PROBE))

        elif not self.units(PYLON).exists and not self.already_pending(PYLON):
            if self.can_afford(PYLON):
                builder = await self.select_worker(nexus.position)
                if builder.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN], builder):
                    await self.build(PYLON, unit=builder, near=self.main_base_ramp.depot_in_middle.towards(nexus.position, 5))
                #await self.build(PYLON, near=nexus.position.random_on_distance(random.randrange(5, 12)))

        if self.supply_left < 5 and not self.already_pending(PYLON) and self.supply_used<190:
            if self.can_afford(PYLON):
                await self.build(PYLON, nexus.position.random_on_distance(random.randrange(1, 2)).towards(self.game_info.map_center, 8))
            return

        if self.units(PYLON).ready.exists:
            pylon = self.units(PYLON).ready.random
            if self.units(GATEWAY).ready.exists:
                if not self.units(CYBERNETICSCORE).exists:
                    if self.can_afford(CYBERNETICSCORE) and not self.already_pending(CYBERNETICSCORE):
                        builder = await self.select_worker(nexus.position)
                        if builder.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN, MOVE], builder):
                            await self.build(CYBERNETICSCORE, near = list(self.main_base_ramp.corner_depots)[0])
            else:
                if self.can_afford(GATEWAY) and not self.already_pending(GATEWAY):
                    builder = await self.select_worker(pylon.position)                    
                    if builder.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN, MOVE], builder):
                        await self.build(GATEWAY, unit=builder, near=list(self.main_base_ramp.corner_depots)[1])
                    #await self.build(GATEWAY, near=pylon)

        for nexus in self.units(NEXUS).ready:
            if self.units(PYLON).ready.exists and not self.already_pending(ASSIMILATOR):
                vgs = self.state.vespene_geyser.closer_than(20.0, nexus)
                for vg in vgs:
                    if not self.can_afford(ASSIMILATOR):
                        break

                    worker = self.select_build_worker(vg.position)
                    if worker is None:
                        break

                    if not self.units(ASSIMILATOR).closer_than(1.0, vg).exists:
                        #await self.do_actions([worker.build(ASSIMILATOR, vg), worker.gather(self.state.mineral_field.closest_to(nexus))])
                        await self.do(worker.build(ASSIMILATOR, vg))
                        await self.do(worker.gather(self.state.mineral_field.closest_to(nexus), queue=True))
                        
            elif self.units(GATEWAY).ready.exists and self.units(ASSIMILATOR).amount < 3:
                vgs = self.state.vespene_geyser.closer_than(20.0, nexus)
                for vg in vgs:
                    if not self.can_afford(ASSIMILATOR):
                        break

                    worker = self.select_build_worker(vg.position)
                    if worker is None:
                        break

                    if not self.units(ASSIMILATOR).closer_than(1.0, vg).exists:
                        await self.do(worker.build(ASSIMILATOR, vg))
                        await self.do(worker.gather(self.state.mineral_field.closest_to(nexus), queue=True))
                        
            elif self.units(CYBERNETICSCORE).ready.exists and self.minerals>500:
                vgs = self.state.vespene_geyser.closer_than(20.0, nexus)
                for vg in vgs:
                    if not self.can_afford(ASSIMILATOR):
                        break

                    worker = self.select_build_worker(vg.position)
                    if worker is None:
                        break

                    if not self.units(ASSIMILATOR).closer_than(1.0, vg).exists:
                        await self.do(worker.build(ASSIMILATOR, vg))
                        await self.do(worker.gather(self.state.mineral_field.closest_to(nexus), queue=True))

        if self.units(PYLON).ready.exists and self.units(CYBERNETICSCORE).ready.exists:
            pylon = self.units(PYLON).ready.random
            if self.units(STARGATE).amount < 2 and not self.already_pending(STARGATE):
                if self.can_afford(STARGATE) and (self.can_afford(VOIDRAY) or self.units(STARGATE).amount < 1):
                    builder = await self.select_worker(pylon.position)
                    if builder.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN, MOVE], builder):
                        await self.build(STARGATE, near=pylon)

        if self.units(PYLON).ready.exists and self.can_afford(PYLON) and self.units(CYBERNETICSCORE).ready.exists and not self.units(PYLON).closer_than(40, self.enemy_start_locations[0]).exists and not self.already_pending(PYLON):
            await self.build(PYLON, near=self.enemy_start_locations[0].towards(self.game_info.map_center, 31))
        
        #print(self.units(PYLON).ready.exists, self.units(CYBERNETICSCORE).ready.exists, not self.units(PYLON).closer_than(40, self.enemy_start_locations[0]).exists, not self.already_pending(PYLON))
        if self.units(PYLON).closer_than(40, self.enemy_start_locations[0]).ready.exists:
            proxy_pylon = self.units(PYLON).closer_than(40, self.enemy_start_locations[0]).random  
            builder = await self.select_worker(proxy_pylon.position)
            if self.units(PYLON).ready.exists and self.units(CYBERNETICSCORE).ready.exists and self.minerals>self.vespene and builder is not None and self.units(SHIELDBATTERY).amount < 10 and self.can_afford(SHIELDBATTERY):
                if builder.is_idle or self.has_order([HARVEST_GATHER, HARVEST_RETURN, MOVE], builder):
                    await self.build(SHIELDBATTERY, unit=builder, near=proxy_pylon)
            
